/*
**
**  A very simple MLP back-propagation driver program for 
**  bunched mode backprop using the fast matrix matrix routines.
**  This code does almost no error checking and could core
**  dump given incorrect command line parameters.
**  The following are hard-coded assumptions:
**     o sigmoid output units
**     o bias units on input and hidden layer fixed as -1 valued.
**     o one-time fixed learning rate
**     o no cross-validation
**     o stop when epoch sum-squared-error goes below threshold.
**
**  $Header: /da/bilmes/phipac/backprop/RCS/mlp.c,v 1.3 1995/12/14 01:52:54 bilmes Exp $
**  Jeff Bilmes <bilmes@icsi.berkeley.edu>
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>

#define DEFBIASUNITVAL (-1.0)
#define TRUE 1
#define FALSE 0

extern double drand48();
extern unsigned short* seed48();

extern float 
bunched_mlp2_backprop (
	  int Ni,     /* Number input units */
	  int Nh,     /* Number hidden units */
	  int No,     /* Number output units */
	  int Np,     /* number patterns in a bunch (i.e., bunch size). */
	  float *who, /* hidden to output weight matrix NoXNh */
	  float *wih, /* input to hidden weight matrix NhXNi */
	  /* for ipats and tpats, there are Np patterns (the bunch size)
	     and each one is a row in the corresponding matrix. */
	  float *ipats, /* input patterns  NpXNi */
	  float *tpats, /* target patterns NpXNo */
	  /* internal temps. preallocated from the outside */
	  float *hpats,  /* hidden layer NpXNh */
	  float *dhids,  /* delta hiddens NpXNh */
	  float *opats,  /* output layer NpXNo */
	  float *douts,  /* delta outputs NpXNo  */
	  float alpha );  /* learning rate */


int ninp;
int nhid;
int nout;
int npat;
int bsize;
char *ifile;
char *tfile;
float alpha;
float bval = DEFBIASUNITVAL;
float esse;
char *ih_file = "/dev/null";
char *ho_file = "/dev/null";
int epr=8;

void
doseed()
{
  int i;
  unsigned short seed16v[3];
  for (i=0;i<3;i++)
     seed16v[i] = time(0);
  seed48(seed16v);
}


void *
lmalloc(int bytes)
{
  void *tmp;
  if ((tmp = (void*)malloc(bytes)) == NULL) {
    fprintf(stderr,"Can't allocate memory\n");
    exit(-1);
  }
  return tmp;
}


void usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  
  fprintf(stderr,"  -ninp i                 : number of input units.\n");
  fprintf(stderr,"  -nhid h                 : number of hidden units.\n");
  fprintf(stderr,"  -nout o                 : number of output units.\n");
  fprintf(stderr,"  -npat p                 : number of training patterns.\n");
  fprintf(stderr,"  -ifile file             : input patterns (must be pXi in size)\n");
  fprintf(stderr,"  -tfile file             : target patterns (must be pXo in size)\n");
  fprintf(stderr,"  -bsize b                : bunch size\n");
  fprintf(stderr,"  -alpha a                : learning rate\n");
  fprintf(stderr,"  -bval f                 : bias unit value\n");
  fprintf(stderr,"  -esse e                 : epoch sum-square-error\n");
  fprintf(stderr,"  -ih_file file           : file to place resulting input->hidden weights\n");
  fprintf(stderr,"  -ho_file file           : file to place resulting hidden->output weights\n");
  fprintf(stderr,"  -epr n                  : epochs per print of sse\n");
}



void
init_matrix(float *mat,int M,int N)
{
  float *endp = &mat[M*N];
  while (mat != endp) {
    *mat++ = drand48();
  }

}



void
bias_matrix(float *mat,int M,int N)
{
  float *p = &mat[N]-1;
  while (M--) {
    *p = bval;
    p += N;
  }
}


void
read_matrix(char *fname,float*mat,int M,int N,int add_bias)
{
  int i,j;
  FILE *file;
  
  if ((file = fopen(fname,"r")) == NULL) {
    fprintf(stderr,"Can't open %s\n",fname);
    exit(-1);
  }
  for (i=0;i<M;i++) {
    for (j=0;j<N;j++) {
      fscanf(file,"%f",mat++);
    }
    if (add_bias)
      *mat++ = bval;
  }
  fclose(file);

}


void
write_matrix(char *fname,float*mat,int M,int N)
{
  int i,j;
  FILE *file;

  if ((file = fopen(fname,"w")) == NULL) {
    fprintf(stderr,"Can't open %s\n",fname);
    exit(-1);
  }

  for (i=0;i<M;i++) {
    for (j=0;j<(N-1);j++) {
      fprintf(file,"%f ",*mat++);
    }
    fprintf(file,"%f\n",*mat++);
  }  
  fclose(file);
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"ninp")) {
	ninp = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"nhid")) {
	nhid = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"nout")) {
	nout = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"npat")) {
	npat = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"bsize")) {
	bsize = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"epr")) {
	epr = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"alpha")) {
	alpha = atof(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"bval")) {
	bval = atof(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"esse")) {
	esse = atof(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"ifile")) {
	ifile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"tfile")) {
	tfile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ih_file")) {
	ih_file = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ho_file")) {
	ho_file = argv[++i];
      } else  {
        fprintf(stderr,"Error: Unknown or invalid argument (%s)\n",argv[i]);
	exit(-1);
      }
    }
  }
  if (ninp <= 0 || nhid <= 0 || nout <= 0 || npat <= 0 ||
      bsize <= 0 || alpha <= 0 || esse <= 0 || ifile == NULL || tfile == NULL ||
      ih_file == NULL || ho_file == NULL) {
    fprintf(stderr,"Error with arguments\n");
    usage(argv[0]);
    exit(-1);
  }

}

int
main(int argc,char *argv[]) 
{
  float J;
  int epoch=0;
  int num;

  /* weights */
  float *who;  /* hidden to output weights */
  float *wih;  /* hidden to output weights */
  
  /* patterns */
  float *ipats,*ipatsp; /* input patterns */
  float *tpats,*tpatsp; /* input patterns */

  /* scratch & misc. */
  float *hpats; /* hidden layer */
  float *dhids; /* delta hiddens */
  float *opats; /* output layer */
  float *douts; /* delta outputs */

  int iters, rem, i, ijump,tjump;
  
  parse_args(argc,argv);
  doseed();

  who = lmalloc(nout*(nhid+1)*sizeof(float));
  wih = lmalloc(nhid*(ninp+1)*sizeof(float));
  ipats = lmalloc(npat*(ninp+1)*sizeof(float));
  tpats = lmalloc(npat*nout*sizeof(float));  
  hpats = lmalloc(bsize*(nhid+1)*sizeof(float));  
  dhids = lmalloc(bsize*nhid*sizeof(float));  
  opats = lmalloc(bsize*nout*sizeof(float));  
  douts = lmalloc(bsize*nout*sizeof(float));  

  read_matrix(ifile,ipats,npat,ninp,TRUE);
  read_matrix(tfile,tpats,npat,nout,FALSE);
  init_matrix(who,nout,(nhid+1));
  init_matrix(wih,nhid,(ninp+1));
  bias_matrix(hpats,bsize,nhid+1);

  iters = npat/bsize;
  rem = npat % bsize;
  ijump = bsize*(ninp+1);
  tjump = bsize*nout;

  J=esse*2;
  epr--;
  num = epr;
  while (J > esse) {
    ipatsp = ipats;
    tpatsp = tpats;

    J=0;
    for (i=0;i<iters;i++) {
      J += bunched_mlp2_backprop(ninp,nhid,nout,bsize,
				 who,wih,
				 ipatsp,tpatsp,
				 hpats,dhids,opats,douts,alpha);
      ipatsp += ijump;
      tpatsp += tjump;
    }
    J += bunched_mlp2_backprop(ninp,nhid,nout,rem,
			       who,wih,
			       ipatsp,tpatsp,
			       hpats,dhids,opats,douts,alpha);
    
    if (num-- == 0) {
      printf("Final esse for epoch %d is %f\n",epoch,J);
      num = epr;
    }
    epoch++;
  }
  printf("Final esse for epoch %d is %f\n",epoch++,J);

  write_matrix(ih_file,wih,nhid,ninp+1);
  write_matrix(ho_file,who,nout,nhid+1);

  
  exit(0);
}
